import cv2
import numpy as np
from scipy.spatial.distance import cdist
from scipy.optimize import linear_sum_assignment


class SimpleAffineTransform:
    def __init__(self, translation=(0, 0), scale=1.0):
        self.translation = np.array(translation)
        self.scale = scale

    def estimate(self, src, dst):
        src_center = np.mean(src, axis=0)
        dst_center = np.mean(dst, axis=0)
        self.translation = dst_center - src_center
        src_dists = np.linalg.norm(src - src_center, axis=1)
        dst_dists = np.linalg.norm(dst - dst_center, axis=1)
        self.scale = np.mean(dst_dists) / (np.mean(src_dists) + 1e-10)

    def inverse(self):
        return SimpleAffineTransform(-self.translation, 1.0 / self.scale)

    def __call__(self, coords):
        return (
            self.scale * (coords - np.mean(coords, axis=0))
            + np.mean(coords, axis=0)
            + self.translation
        )

    def residuals(self, src, dst):
        return np.sqrt(np.sum((self(src) - dst) ** 2, axis=1))


def norm_coords(x, left, right):
    if x < left:
        return left
    if x > right:
        return right
    return x


def norm_same_token(token):
    special_map = {
        "\\dot": ".",
        "\\Dot": ".",
        "\\cdot": ".",
        "\\cdotp": ".",
        "\\ldotp": ".",
        "\\mid": "|",
        "\\rightarrow": "\\to",
        "\\top": "T",
        "\\Tilde": "\\tilde",
        "\\prime": "'",
        "\\ast": "*",
        "\\left<": "\\langle",
        "\\right>": "\\rangle",
        "\\lbrace": "\{",
        "\\rbrace": "\}",
        "\\lbrack": "[",
        "\\rbrack": "]",
        "\\blackslash": "/",
        "\\slash": "/",
        "\\leq": "\\le",
        "\\geq": "\\ge",
        "\\neq": "\\ne",
        "\\Vert": "\\|",
        "\\lVert": "\\|",
        "\\rVert": "\\|",
        "\\vert": "|",
        "\\lvert": "|",
        "\\rvert": "|",
        "\\colon": ":",
        "\\Ddot": "\\ddot",
        "\\Bar": "\\bar",
        "\\Vec": "\\vec",
        "\\parallel": "\\|",
        "\\dag": "\\dagger",
        "\\ddag": "\\ddagger",
        "\\textlangle": "<",
        "\\textrangle": ">",
        "\\textgreater": ">",
        "\\textless": "<",
        "\\textbackslash": "\\",
        "\\textunderscore": "_",
        "\\=": "=",
        "\\neg": "\\lnot",
        "\\neq": "\\not=",
    }
    if token.startswith("\\left") or token.startswith("\\right"):
        if (
            "arrow" not in token
            and "<" not in token
            and ">" not in token
            and "harpoon" not in token
        ):
            token = token.replace("\\left", "").replace("\\right", "")
    if token.startswith("\\big") or token.startswith("\\Big"):
        if "\\" in token[4:]:
            token = "\\" + token[4:].split("\\")[-1]
        else:
            token = token[-1]
    if token in special_map.keys():
        token = special_map[token]
    if token.startswith("\\wide"):
        return token.replace("wide", "")
    if token.startswith("\\var"):
        return token.replace("var", "")
    if token.startswith("\\string"):
        return token.replace("\\string", "")
    return token


class HungarianMatcher:
    def __init__(
        self,
        cost_token: float = 1,
        cost_position: float = 0.05,
        cost_order: float = 0.15,
    ):
        self.cost_token = cost_token
        self.cost_position = cost_position
        self.cost_order = cost_order
        self.cost = {}

    def calculate_token_cost(self, box_gt, box_pred):

        all_tokens = [data["token"] for data in box_gt + box_pred]
        unique_tokens = sorted(list(set(all_tokens)))
        token2id = {token: i for i, token in enumerate(unique_tokens)}
        num_classes = len(token2id)

        all_norm_tokens = [norm_same_token(data["token"]) for data in box_gt + box_pred]
        unique_norm_tokens = sorted(list(set(all_norm_tokens)))
        token2id_norm = {token: i for i, token in enumerate(unique_norm_tokens)}
        num_classes_norm = len(token2id_norm)

        gt_token_array = np.array([token2id[data["token"]] for data in box_gt])
        norm_gt_token_array = np.array(
            [token2id_norm[norm_same_token(data["token"])] for data in box_gt]
        )

        pred_token_logits = np.zeros((len(box_pred), num_classes))
        for i, data in enumerate(box_pred):
            if data["token"] in token2id:
                pred_token_logits[i, token2id[data["token"]]] = 1

        norm_pred_token_logits = np.zeros((len(box_pred), num_classes_norm))
        for i, data in enumerate(box_pred):
            norm_token = norm_same_token(data["token"])
            if norm_token in token2id_norm:
                norm_pred_token_logits[i, token2id_norm[norm_token]] = 1

        if gt_token_array.size == 0 or pred_token_logits.shape[0] == 0:
            return np.empty((len(box_gt), len(box_pred)))

        token_cost = 1.0 - pred_token_logits[:, gt_token_array]
        norm_token_cost = 1.0 - norm_pred_token_logits[:, norm_gt_token_array]

        token_cost[np.logical_and(token_cost == 1, norm_token_cost == 0)] = 0.005
        return token_cost.T

    def box2array(self, box_list, size):
        W, H = size
        box_array = []
        for box in box_list:
            x_min, y_min, x_max, y_max = box["bbox"]
            box_array.append(
                [
                    (x_min + x_max) / (2 * W),
                    (y_min + y_max) / (2 * H),
                    (x_max - x_min) / W,
                    (y_max - y_min) / H,
                ]
            )
        return np.array(box_array)

    def order2array(self, box_list, max_token_lens=None):
        if not max_token_lens:
            max_token_lens = len(box_list)
        return np.array([[idx / max_token_lens] for idx, _ in enumerate(box_list)])

    def calculate_l1_cost(self, gt_array, pred_array):
        if gt_array.shape[0] == 0 or pred_array.shape[0] == 0:
            return np.empty((gt_array.shape[0], pred_array.shape[0]))
        return cdist(gt_array, pred_array, "minkowski", p=1) / gt_array.shape[-1]

    def __call__(self, box_gt, box_pred, gt_size, pred_size):
        if not box_gt or not box_pred:
            return []
        gt_box_array = self.box2array(box_gt, gt_size)
        pred_box_array = self.box2array(box_pred, pred_size)
        max_token_lens = max(len(box_gt), len(box_pred))
        gt_order_array = self.order2array(box_gt, max_token_lens)
        pred_order_array = self.order2array(box_pred, max_token_lens)
        token_cost = self.calculate_token_cost(box_gt, box_pred)
        position_cost = self.calculate_l1_cost(gt_box_array, pred_box_array)
        order_cost = self.calculate_l1_cost(gt_order_array, pred_order_array)
        self.cost = {
            "token": token_cost,
            "position": position_cost,
            "order": order_cost,
        }
        cost = (
            self.cost_token * token_cost
            + self.cost_position * position_cost
            + self.cost_order * order_cost
        )
        cost[np.isnan(cost) | np.isinf(cost)] = 100
        row_ind, col_ind = linear_sum_assignment(cost)
        return list(zip(row_ind, col_ind))


def update_inliers(ori_inliers, sub_inliers):
    inliers = np.copy(ori_inliers)
    sub_idx = -1
    for idx in range(len(ori_inliers)):
        if ori_inliers[idx] == False:
            sub_idx += 1
            if sub_inliers[sub_idx] == True:
                inliers[idx] = True
    return inliers




import re
import subprocess
from pathlib import Path
from typing import Tuple
import sys


IMPLICIT_MULTIPLICATION_TARGETS = [
    "arccos",
    "arcsin",
    "arctan",
    "sinh",
    "cosh",
    "tanh",
    "coth",
    "sech",
    "csch",
    "lim",
    "log",
    "ln",
    "exp",
    "sin",
    "cos",
    "tan",
    "cot",
    "sec",
    "csc",
    "ArcCos",
    "ArcSin",
    "ArcTan",
    "Sinh",
    "Cosh",
    "Tanh",
    "Coth",
    "Sech",
    "Csch",
    "Lim",
    "Log",
    "Ln",
    "Exp",
    "Sin",
    "Cos",
    "Tan",
    "Cot",
    "Sec",
    "Csc",
    "det",
    "dim",
    "min",
    "max",
    "sup",
    "inf",
    "deg",
    "gcd",
    "lcm",
    "ker",
    "im",
    "Pr",
    "E",
    "Var",
    "Cov",
    "rank",
    "Tr",
    "span",
    "proj",
    "grad",
    "div",
    "curl",
    "Res",
    "pi",
    "alpha",
    "beta",
    "gamma",
    "delta",
    "epsilon",
    "zeta",
    "eta",
    "theta",
    "iota",
    "kappa",
    "lambda",
    "mu",
    "nu",
    "xi",
    "omicron",
    "rho",
    "sigma",
    "tau",
    "upsilon",
    "phi",
    "chi",
    "psi",
    "omega",
    "Pi",
    "Alpha",
    "Beta",
    "Gamma",
    "Delta",
    "Epsilon",
    "Zeta",
    "Eta",
    "Theta",
    "Iota",
    "Kappa",
    "Lambda",
    "Mu",
    "Nu",
    "Xi",
    "Omicron",
    "Rho",
    "Sigma",
    "Tau",
    "Upsilon",
    "Phi",
    "Chi",
    "Psi",
    "Omega",
    "varepsilon",
    "vartheta",
    "varpi",
    "varrho",
    "varsigma",
    "varphi",
    "partial",
    "nabla",
    "int",
    "oint",
    "sum",
    "prod",
    "wedge",
    "vee",
    "cap",
    "cup",
    "therefore",
    "because",
    "Rightarrow",
    "rightarrow",
    "Leftarrow",
    "leftarrow",
    "Leftrightarrow",
    "leftrightarrow",
    "in",
    "ni",
    "subset",
    "supset",
    "subseteq",
    "supseteq",
    "equiv",
    "sim",
    "simeq",
    "approx",
    "propto",
    "cdot",
    "times",
    "otimes",
    "oplus",
    "quad",
    "qquad",
    "ldots",
    "cdots",
    "ddots",
    "forall",
    "exists",
    "neg",
    "infty",
    "hbar",
    "emptyset",
    "angle",
    "geqslant",
    "hat",
    "bar",
    "tilde",
    "vec",
    "dot",
    "ddot",
    "sqrt",
]
IMPLICIT_MULTIPLICATION_TARGETS.sort(key=len, reverse=True)
TARGETS_PATTERN = "|".join(IMPLICIT_MULTIPLICATION_TARGETS)

PATTERN_ALIGN_ENV = re.compile(
    r"\\begin{(split|align|alignedat|alignat|eqnarray)\*?}(.+?)\\end{\1\*?}", re.S
)
PATTERN_SMALLMATRIX = re.compile(r"\\begin{(smallmatrix)\*?}(.+?)\\end{\1\*?}", re.S)
PATTERN_INVALID_SINGLE_CHAR_CMD = re.compile(r"\\([a-zA-Z0-9])(?![a-zA-Z])")
PATTERN_LATEX_CMD_CONCAT_CMD = re.compile(
    r"\\(" + TARGETS_PATTERN + r")" + r"(\\[a-zA-Z])"
)
PATTERN_LATEX_CMD_CONCAT_TEXT = re.compile(r"\\(" + TARGETS_PATTERN + r")([a-zA-Z])")
PATTERN_NON_CMD_IMPLICIT_MULT = re.compile(
    r"\b(" + TARGETS_PATTERN + r")([a-zA-Z][a-zA-Z0-9]*)\b"
)

OPERATORS = "\s?".join(
    "|".join(
        [
            "arccos",
            "arcsin",
            "arctan",
            "arg",
            "cos",
            "cosh",
            "cot",
            "coth",
            "csc",
            "deg",
            "det",
            "dim",
            "exp",
            "gcd",
            "hom",
            "inf",
            "injlim",
            "ker",
            "lg",
            "lim",
            "liminf",
            "limsup",
            "ln",
            "log",
            "max",
            "min",
            "Pr",
            "projlim",
            "sec",
            "sin",
            "sinh",
            "sup",
            "tan",
            "tanh",
        ]
    )
)
PATTERN_OPERATOR_NAME = re.compile(r"\\operatorname {(%s)}" % OPERATORS)



def tokenize(latex_code: str) -> Tuple[bool, str]:

    if not latex_code:
        return True, ""

    root_dir = Path(__file__).parent
    formula_script = root_dir / "tokenize_latex" / "preprocess_formula.js"

    prepre = latex_code.replace("\n", " ")
    prepre = PATTERN_ALIGN_ENV.sub(r"\\begin{aligned}\2\\end{aligned}", prepre)
    prepre = PATTERN_SMALLMATRIX.sub(r"\\begin{matrix}\2\\end{matrix}", prepre)
    prepre = PATTERN_INVALID_SINGLE_CHAR_CMD.sub(r"\1", prepre)
    prepre = PATTERN_LATEX_CMD_CONCAT_CMD.sub(r"\\\1 \2", prepre)
    prepre = PATTERN_LATEX_CMD_CONCAT_TEXT.sub(r"\\\1 \2", prepre)
    prepre = PATTERN_NON_CMD_IMPLICIT_MULT.sub(r"\1 \2", prepre)

    try:
        proc = subprocess.run(
            ["node", str(formula_script), "normalize"],
            input=prepre,
            capture_output=True,
            text=True,
            check=True,
            encoding="utf-8",
        )
        normalized_latex = proc.stdout
    except (subprocess.CalledProcessError, FileNotFoundError) as e:
        print(f"Error executing Node.js script (formula): {e}", file=sys.stderr)
        if hasattr(e, "stderr"):
            print(f"Node.js stderr: {e.stderr}", file=sys.stderr)
        return False, latex_code

    names = [
        "\\" + x.replace(" ", "")
        for x in re.findall(PATTERN_OPERATOR_NAME, normalized_latex)
    ]
    post = PATTERN_OPERATOR_NAME.sub(
        lambda match: str(names.pop(0)), normalized_latex
    ).replace(r"\\ \end{array}", r"\end{array}")
    return True, post.strip()



import colorsys
import re
import sys
from typing import List, Tuple
from fastcdm.latex_processor import normalize_latex, token_add_color_RGB


def generate_high_contrast_colors(num_colors: int) -> List[Tuple[int, int, int]]:
    """
    使用 HSV 颜色空间生成一系列高对比度、视觉上易于区分的颜色。
    """
    colors_rgb = []
    golden_ratio_conjugate = 0.61803398875
    hue = 0.5
    for _ in range(num_colors):
        hue += golden_ratio_conjugate
        hue %= 1
        rgb_float = colorsys.hsv_to_rgb(hue, 0.9, 0.95)
        rgb_int = tuple(int(c * 255) for c in rgb_float)
        colors_rgb.append(rgb_int)
    return colors_rgb


def process_for_katex(pre_tokenized_latex: str) -> Tuple[str, List[str]]:

    if not pre_tokenized_latex or not isinstance(pre_tokenized_latex, str):
        return "", []

    try:
        normalized_latex = normalize_latex(pre_tokenized_latex)
        l_split = [token for token in normalized_latex.strip().split(" ") if token]

        idx = 0
        token_list = []
        temp_l_split = list(l_split)

        while idx < len(temp_l_split):
            temp_l_split, idx, token_list = token_add_color_RGB(
                temp_l_split, idx, token_list
            )

        colored_template_original = " ".join(temp_l_split)
        katex_template = re.sub(
            r"\\mathcolor\[RGB\]{<color_(\d+)>}",
            r"\\color{__COLOR__\1__}",
            colored_template_original,
        )
        return katex_template, token_list

    except Exception as e:
        print(f"\n在 KaTeX 处理过程中出错: {e}", file=sys.stderr)
        print(f"有问题的 LaTeX 字符串: {pre_tokenized_latex}", file=sys.stderr)
        return "[PROCESSING FAILED]", []



import re

# 以下列表定义了在后续 token_add_color 系列函数中“跳过”的正则模式。
# 任何匹配这些模式的 token 都不会被着色，而是保持原样（通常用黑色标记）。
# 主要用于括号、环境边界、上下标等结构元素。
SKIP_PATTERNS = [
    r"\{",
    r"\}",
    r"[\[\]]",
    r"\\begin\{.*?\}",
    r"\\end\{.*?\}",
    r"\^",
    r"\_",
    r"\\.*rule.*",
    r"\\.*line.*",
    r"\[[\-.0-9]+[epm][xtm]\]",
]

# 以下列表中的 LaTeX 命令在后续处理中被视为“透明”或“无意义”的 token。
# 它们不会触发着色逻辑，直接跳过，避免干扰真正的数学内容。
SKIP_Tokens = [
    "\\",
    "\\\\",
    "\\index",
    "\\a",
    "&",
    "$",
    "\\multirow",
    "\\def",
    "\\edef",
    "\\raggedright",
    "\\url",
    "\\cr",
    "\\ensuremath",
    "\\left",
    "\\left[",
    "\\left(",
    "\\left{",
    "\\right",
    "\\right]",
    "\\right)",
    "\\right}",
    "\\mathchoice",
    "\\scriptstyle",
    "\\displaystyle",
    "\\qquad",
    "\\quad",
    "\\,",
    "\\!",
    "~",
    "\\boldmath",
    "\\gdef",
    "\\today",
    "\\the",
]

# PHANTOM_Tokens 中的命令在后续着色时被视为“幻影”命令：
# 它们本身不直接参与颜色标记，但其参数仍需递归处理。
# 主要用于字体、引用、颜色等不影响数学结构的命令。
PHANTOM_Tokens = [
    "\\fontfamily",
    "\\vphantom",
    "\\phantom",
    "\\rowcolor",
    "\\ref",
    "\\thesubequation",
    "\\global",
    "\\theboldgroup",
]

# 以下命令在后续处理中被识别为“双尾”命令：它们需要两个 {} 参数。
# 例如 \frac{分子}{分母}，在着色时会分别对两个参数进行灰色处理。
TWO_Tail_Tokens = ["\\frac", "\\binom"]

# AB_Tail_Tokens 中的命令具有“可选+必选”参数结构：
# 第一个参数可以是 []，第二个必须是 {}。
# 例如 \xrightarrow[下方]{上方}，在着色时会分别处理两个参数。
AB_Tail_Tokens = ["\\xrightarrow", "\\xleftarrow", "\\sqrt"]  # special token \xxx [] {}

# 以下命令也是“双尾”但被视为“不可见”结构，着色逻辑与 TWO_Tail_Tokens 类似，
# 但通常用于上下堆叠等排版，不影响数学含义。
TWO_Tail_Invisb_Tokens = ["\\overset", "\\underset", "\\stackrel"]

# ONE_Tail_Tokens 中的命令只需一个 {} 参数，且会显著改变数学符号外观。
# 在着色时，命令本身保持黑色，参数内容置灰。
ONE_Tail_Tokens = [
    "\\widetilde",
    "\\overline",
    "\\hat",
    "\\widehat",
    "\\tilde",
    "\\Tilde",
    "\\dot",
    "\\bar",
    "\\vec",
    "\\underline",
    "\\underbrace",
    "\\check",
    "\\breve",
    "\\Bar",
    "\\Vec",
    "\\mathring",
    "\\ddot",
    "\\Ddot",
    "\\dddot",
    "\\ddddot",
]

# ONE_Tail_Invisb_Tokens 中的命令同样只需一个 {} 参数，
# 但主要用于字体或样式切换，不改变数学含义，因此整体视为“不可见”，
# 在着色时命令本身不标记，仅对其参数递归处理。
ONE_Tail_Invisb_Tokens = [
    "\\boldsymbol",
    "\\pmb",
    "\\textbf",
    "\\mathrm",
    "\\mathbf",
    "\\mathsf",
    "\\mathbb",
    "\\mathcal",
    "\\mathinner",
    "\\mathit",
    "\\mathnormal",
    "\\mathring",
    "\\mathscr",
    "\\mathtt",
    "\\textmd",
    "\\texttt",
    "\\textnormal",
    "\\text",
    "\\textit",
    "\\textup",
    "\\mathop",
    "\\mathbin",
    "\\smash",
    "\\operatorname",
    "\\textrm",
    "\\mathfrak",
    "\\emph",
    "\\textsf",
    "\\textsc",
]


def flatten_multiline(latex):
    brace_map = {
        "\\left(": "\\right)",
        "\\left[": "\\right]",
        "\\left{": "\\right}",
    }
    l_split = latex.split(" ")
    if l_split[0] == "\\begin{array}":
        if l_split[-1] == "\\end{array}":
            l_split = l_split[2:-1]
        else:
            l_split = l_split[2:]

    idx = 0
    while idx < len(l_split):
        token = l_split[idx]
        if token.startswith("\\left") and token in brace_map.keys():
            end_idx = find_matching_brace(l_split, idx, brace=[token, brace_map[token]])
            if end_idx != -1:
                idx = end_idx
        elif token in ["\\\\", "~", "\\qquad"]:
            l_split = l_split[0:idx] + l_split[idx + 1 :]
            idx -= 1
        idx += 1
    latex = " ".join(l_split)
    return "$ " + latex + " $"


def clean_latex(text):
    cleaned_text = re.sub(r"(?<=[^\\])\s+(?=[^\\])", "", text)
    for item in [
        "\\hline",
        "\\midrule",
        "\\times",
        "\\bf",
        "\\footnotesize",
        "\\cr",
        "\\log",
    ]:
        cleaned_text = cleaned_text.replace(item, item + " ")
    cleaned_text = cleaned_text.replace(" \\mathcolor{black}", "\\mathcolor{black}")
    return cleaned_text


def remove_trailing_latex(formula):
    pattern = r"(\\(hspace\*?\{[^{}]*?\}|vspace\*?\{[^{}]*?\}|smallskip|medskip|quad|qquad|bigskip|[;,])|\~|\.)*$"
    cleaned_formula = re.sub(pattern, "", formula, count=1)
    return cleaned_formula


def find_matching_brace(sequence, start_index, brace=["{", "}"]):
    left_brace, right_brace = brace
    depth = 0
    for i, char in enumerate(sequence[start_index:], start=start_index):
        if char == left_brace:
            depth += 1
        elif char == right_brace:
            depth -= 1
            if depth == 0:
                return i
    if depth > 0:
        error_info = "Warning! found no matching brace in sequence !"
        raise ValueError(error_info)
    return -1


def normalize_latex(l, rm_trail=False):
    if "tabular" in l:
        latex_type = "tabular"
    else:
        latex_type = "formula"

    if rm_trail:
        l = remove_trailing_latex(l)
    l = l.strip().replace(r"\pmatrix", r"\mypmatrix").replace(r"\matrix", r"\mymatrix")

    for item in ["\\raggedright", "\\arraybackslash"]:
        l = l.replace(item, "")

    for item in ["\\lowercase", "\\uppercase"]:
        l = l.replace(item, "")

    pattern = r"\\[hv]space { [.0-9a-z ]+ }"
    old_token = re.findall(pattern, l, re.DOTALL)
    if latex_type == "tabular":
        new_token = ["" for item in old_token]
    else:
        new_token = [item.replace(" ", "") for item in old_token]
    for bef, aft in zip(old_token, new_token):
        l = l.replace(bef, aft)

    if latex_type == "tabular":
        l = l.replace("\\begin {tabular}", "\\begin{tabular}")
        l = l.replace("\\end {tabular}", "\\end{tabular}")
        l = l.replace("\\begin {array}", "\\begin{array}")
        l = l.replace("\\end {array}", "\\end{array}")
        l_split = l.split(" ")
        idx = 0
        while idx < len(l_split):
            token = l_split[idx]
            if token == "\\begin{tabular}":
                sub_idx = idx + 1
                end_idx = find_matching_brace(l_split, sub_idx)
                new_token = "".join(l_split[idx : end_idx + 1])
                l_split = l_split[0:idx] + [new_token] + l_split[end_idx + 1 :]
                break
            idx += 1
        l = " ".join(l_split)
        l_split = l.split(" ")
        idx = 0
        while idx < len(l_split):
            token = l_split[idx]
            if token in ["\\cmidrule", "\\cline"]:
                sub_idx = idx + 1
                if l_split[sub_idx] == "(":
                    mid_end = find_matching_brace(l_split, sub_idx, brace=["(", ")"])
                    end_idx = find_matching_brace(l_split, mid_end + 1)
                else:
                    end_idx = find_matching_brace(l_split, sub_idx)
                new_token = "".join(l_split[idx : end_idx + 1])
                l_split = l_split[0:idx] + [new_token] + l_split[end_idx + 1 :]
            idx += 1
        l = " ".join(l_split)

    pattern = r"\\begin{array} { [lrc ]+ }"
    old_token = re.findall(pattern, l, re.DOTALL)
    new_token = [
        item.replace("\\begin{array} ", "<s>")
        .replace(" ", "")
        .replace("<s>", "\\begin{array} ")
        for item in old_token
    ]
    for bef, aft in zip(old_token, new_token):
        l = l.replace(bef, aft)

    l = " " + l + " "
    l = re.sub(r"(?<=\s)--(?=\s)", r"- -", l)
    l = re.sub(r"(?<=\s)---(?=\s)", r"- - -", l)
    l = re.sub(r"(?<=\s)…(?=\s)", r". . .", l)
    l = re.sub(r"(?<=\s)\\ldots(?=\s)", r". . .", l)
    l = re.sub(r"(?<=\s)\\hdots(?=\s)", r". . .", l)
    l = re.sub(r"(?<=\s)\\cdots(?=\s)", r". . .", l)
    l = re.sub(r"(?<=\s)\\dddot(?=\s)", r". . .", l)
    l = re.sub(r"(?<=\s)\\dots(?=\s)", r". . .", l)
    l = re.sub(r"(?<=\s)\\dotsc(?=\s)", r". . .", l)
    l = re.sub(r"(?<=\s)\\dotsi(?=\s)", r". . .", l)
    l = re.sub(r"(?<=\s)\\dotsm(?=\s)", r". . .", l)
    l = re.sub(r"(?<=\s)\\dotso(?=\s)", r". . .", l)
    l = re.sub(r"(?<=\s)\\dotsb(?=\s)", r". . .", l)
    l = re.sub(r"(?<=\s)\\mathellipsis(?=\s)", r". . .", l)
    l = re.sub(r"(?<=\s)\\ex(?=\s)", r"\\mathrm { e x }", l)
    l = re.sub(r"(?<=\s)\\ln(?=\s)", r"\\mathrm { l n }", l)
    l = re.sub(r"(?<=\s)\\lg(?=\s)", r"\\mathrm { l g }", l)
    l = re.sub(r"(?<=\s)\\cot(?=\s)", r"\\mathrm { c o t }", l)
    l = re.sub(r"(?<=\s)\\mod(?=\s)", r"\\mathrm { m o d }", l)
    l = re.sub(r"(?<=\s)\\bmod(?=\s)", r"\\mathrm { m o d }", l)
    l = re.sub(r"(?<=\s)\\pmod(?=\s)", r"\\mathrm { m o d }", l)
    l = re.sub(r"(?<=\s)\\min(?=\s)", r"\\mathrm { m i n }", l)
    l = re.sub(r"(?<=\s)\\max(?=\s)", r"\\mathrm { m a x }", l)
    l = re.sub(r"(?<=\s)\\ker(?=\s)", r"\\mathrm { k e r }", l)
    l = re.sub(r"(?<=\s)\\hom(?=\s)", r"\\mathrm { h o m }", l)
    l = re.sub(r"(?<=\s)\\sec(?=\s)", r"\\mathrm { s e c }", l)
    l = re.sub(r"(?<=\s)\\scs(?=\s)", r"\\mathrm { s c s }", l)
    l = re.sub(r"(?<=\s)\\csc(?=\s)", r"\\mathrm { c s c }", l)
    l = re.sub(r"(?<=\s)\\deg(?=\s)", r"\\mathrm { d e g }", l)
    l = re.sub(r"(?<=\s)\\arg(?=\s)", r"\\mathrm { a r g }", l)
    l = re.sub(r"(?<=\s)\\log(?=\s)", r"\\mathrm { l o g }", l)
    l = re.sub(r"(?<=\s)\\dim(?=\s)", r"\\mathrm { d i m }", l)
    l = re.sub(r"(?<=\s)\\exp(?=\s)", r"\\mathrm { e x p }", l)
    l = re.sub(r"(?<=\s)\\sin(?=\s)", r"\\mathrm { s i n }", l)
    l = re.sub(r"(?<=\s)\\cos(?=\s)", r"\\mathrm { c o s }", l)
    l = re.sub(r"(?<=\s)\\tan(?=\s)", r"\\mathrm { t a n }", l)
    l = re.sub(r"(?<=\s)\\tanh(?=\s)", r"\\mathrm { t a n h }", l)
    l = re.sub(r"(?<=\s)\\cosh(?=\s)", r"\\mathrm { c o s h }", l)
    l = re.sub(r"(?<=\s)\\sinh(?=\s)", r"\\mathrm { s i n h }", l)
    l = re.sub(r"(?<=\s)\\coth(?=\s)", r"\\mathrm { c o t h }", l)
    l = re.sub(r"(?<=\s)\\arcsin(?=\s)", r"\\mathrm { a r c s i n }", l)
    l = re.sub(r"(?<=\s)\\arccos(?=\s)", r"\\mathrm { a r c c o s }", l)
    l = re.sub(r"(?<=\s)\\arctan(?=\s)", r"\\mathrm { a r c t a n }", l)
    l = re.sub(r"(?<=\s)\\bf([a-zA-Z])", r"\\mathbf{\1}", l)

    pattern = r"\\string [^ ]+ "
    old_token = re.findall(pattern, l, re.DOTALL)
    new_token = [item.replace(" ", "") for item in old_token]
    for bef, aft in zip(old_token, new_token):
        l = l.replace(bef, aft + " ")

    pattern = r"\\[Bb]ig[g]?[glrm]? [(){}|\[\]] "
    old_token = re.findall(pattern, l, re.DOTALL)
    new_token = [item.replace(" ", "") for item in old_token]
    for bef, aft in zip(old_token, new_token):
        l = l.replace(bef, aft + " ")

    pattern = r"\\[Bb]ig[g]?[glrm]? \\.*? "
    old_token = re.findall(pattern, l, re.DOTALL)
    new_token = [item.replace(" ", "") for item in old_token]
    for bef, aft in zip(old_token, new_token):
        l = l.replace(bef, aft + " ")

    pattern = r"\\operatorname \*"
    old_token = re.findall(pattern, l, re.DOTALL)
    new_token = ["\\operatorname" for item in old_token]
    for bef, aft in zip(old_token, new_token):
        l = l.replace(bef, aft)

    l = l.replace("\\lefteqn", "")

    l = l.replace("\\footnote ", "^ ")

    pattern = r"\\\' [^{] "
    old_token = re.findall(pattern, l, re.DOTALL)
    new_token = [item.replace(" ", "") for item in old_token]
    for bef, aft in zip(old_token, new_token):
        l = l.replace(bef, aft + " ")

    if latex_type == "tabular":
        pattern = r"\[ [\-.0-9 ]+[exptcm ]+ \]"
        old_token = re.findall(pattern, l, re.DOTALL)
        new_token = [item.replace(" ", "") for item in old_token]
        for bef, aft in zip(old_token, new_token):
            l = l.replace(bef, aft)

    pattern = r"\\parbox {[^{]+}"
    old_token = re.findall(pattern, l, re.DOTALL)
    new_token = [item.replace(" ", "") for item in old_token]
    for bef, aft in zip(old_token, new_token):
        l = l.replace(bef, aft)

    pattern = r"\\raisebox {[^{]+} [\[\]0-9 exptcm]+{"
    old_token = re.findall(pattern, l, re.DOTALL)
    new_token = [item.replace(" ", "") for item in old_token]
    for bef, aft in zip(old_token, new_token):
        l = l.replace(bef, aft[0:-1] + " {")

    pattern = r"{ \\char[0-9\' ]+}"
    old_token = re.findall(pattern, l, re.DOTALL)
    new_token = [item.replace(" ", "") for item in old_token]
    for bef, aft in zip(old_token, new_token):
        l = l.replace(bef, "{ " + aft[1:-1] + " }")

    pattern = r"\\rule {[ .0-9a-z]+} {[ .0-9a-z]+}"
    old_token = re.findall(pattern, l, re.DOTALL)
    new_token = [item.replace(" ", "") for item in old_token]
    for bef, aft in zip(old_token, new_token):
        l = l.replace(bef, aft)

    pattern = r"\\specialrule {[ .0-9a-z]+} {[ .0-9a-z]+} {[ .0-9a-z]+}"
    old_token = re.findall(pattern, l, re.DOTALL)
    new_token = [item.replace(" ", "") for item in old_token]
    for bef, aft in zip(old_token, new_token):
        l = l.replace(bef, aft)

    pattern = r"\\colorbox[ \[\]RGBrgb]+{ [A-Za-z 0-9,!]+ } |\\color[ \[\]RGBrgb]+{ [A-Za-z 0-9,!]+ } |\\textcolor[ \[\]RGBrgb]+{ [A-Za-z 0-9,!]+ } |\\cellcolor[ \[\]RGBrgb]+{ [A-Za-z 0-9,!]+ } "
    old_token = re.findall(pattern, l, re.DOTALL)
    for bef in old_token:
        l = l.replace(bef, "")

    l_split = l.split(" ")
    idx = 0
    while idx < len(l_split):
        token = l_split[idx]
        if token in ONE_Tail_Tokens + ONE_Tail_Invisb_Tokens:
            sub_idx = idx + 1
            while (
                sub_idx < len(l_split)
                and l_split[sub_idx] in ONE_Tail_Tokens + ONE_Tail_Invisb_Tokens
            ):
                sub_idx += 1
            new_split = l_split[0:idx]
            for ii in range(idx, sub_idx):
                new_split = new_split + [l_split[ii], "{"]
            if l_split[sub_idx] != "{":
                new_split = new_split + [l_split[sub_idx]] + ["}"] * (sub_idx - idx)
                l_split = new_split + l_split[sub_idx + 1 :]
            else:
                end_idx = find_matching_brace(l_split, sub_idx)
                new_split = (
                    new_split + l_split[sub_idx + 1 : end_idx] + ["}"] * (sub_idx - idx)
                )
                l_split = new_split + l_split[end_idx + 1 :]
        elif token in AB_Tail_Tokens:
            if l_split[idx + 1] != "[" and l_split[idx + 1] != "{":
                l_split = (
                    l_split[0 : idx + 1]
                    + ["{"]
                    + [l_split[idx + 1]]
                    + ["}"]
                    + l_split[idx + 2 :]
                )
            else:
                if l_split[idx + 1] == "[":
                    end1 = find_matching_brace(l_split, idx + 1, brace=["[", "]"])
                else:
                    end1 = idx
                if l_split[end1 + 1] != "{":
                    l_split = (
                        l_split[0 : end1 + 1]
                        + ["{"]
                        + [l_split[end1 + 1]]
                        + ["}"]
                        + l_split[end1 + 2 :]
                    )
        elif token in TWO_Tail_Tokens + TWO_Tail_Invisb_Tokens:
            if l_split[idx + 1] != "{":
                l_split = (
                    l_split[0 : idx + 1]
                    + ["{"]
                    + [l_split[idx + 1]]
                    + ["}"]
                    + l_split[idx + 2 :]
                )
            end1 = find_matching_brace(l_split, idx + 1)
            if l_split[end1 + 1] != "{":
                l_split = (
                    l_split[0 : end1 + 1]
                    + ["{"]
                    + [l_split[end1 + 1]]
                    + ["}"]
                    + l_split[end1 + 2 :]
                )

        idx += 1
    l = " ".join(l_split)

    return l


def token_add_color(l_split, idx, render_dict):
    token = l_split[idx]
    if token in PHANTOM_Tokens:
        if l_split[idx + 1] == "{":
            brace_end = find_matching_brace(l_split, idx + 1)
        else:
            brace_end = idx + 1
        next_idx = brace_end + 1
    elif token in TWO_Tail_Tokens:
        num_start = idx + 1
        num_end = find_matching_brace(l_split, num_start)
        den_start = num_end + 1
        den_end = find_matching_brace(l_split, den_start)
        l_split_copy = (
            l_split[:idx]
            + [r"\mathcolor{black}{" + token + "{"]
            + [r"\mathcolor{gray}{"]
            + l_split[num_start + 1 : num_end]
            + ["}"]
            + [r"}{"]
            + [r"\mathcolor{gray}{"]
            + l_split[den_start + 1 : den_end]
            + ["}"]
            + ["}"]
            + ["}"]
            + l_split[den_end + 1 :]
        )

        l_new = " ".join(l_split_copy)
        l_new = r"\mathcolor{gray}{ " + l_new + " }"
        render_dict[str(idx)] = l_new, token
        next_idx = idx + 1
    elif token in ONE_Tail_Tokens:
        num_start = idx + 1
        num_end = find_matching_brace(l_split, num_start)
        l_split_copy = (
            l_split[:idx]
            + [r"\mathcolor{black}{"]
            + l_split[idx : num_start + 1]
            + [r"\mathcolor{gray}{"]
            + l_split[num_start + 1 : num_end]
            + ["}"]
            + l_split[num_end : num_end + 1]
            + ["}"]
            + l_split[num_end + 1 :]
        )
        l_new = " ".join(l_split_copy)
        l_new = r"\mathcolor{gray}{ " + l_new + " }"
        render_dict[str(idx)] = l_new, token
        next_idx = idx + 1
    elif token in ONE_Tail_Invisb_Tokens:
        num_start = idx + 1
        num_end = find_matching_brace(l_split, num_start)
        sub_idx = num_start + 1
        if num_end - num_start == 2:
            l_split_copy = l_split.copy()
            l_split_copy[sub_idx] = (
                r"{\mathcolor{black}{" + l_split_copy[sub_idx] + "}}"
            )
            l_new = " ".join(l_split_copy)
            l_new = r"\mathcolor{gray}{ " + l_new + " }"
            render_dict[str(idx)] = l_new, l_split[sub_idx]
            next_idx = num_end
        else:
            while sub_idx < num_end:
                l_split, sub_idx, render_dict = token_add_color(
                    l_split, sub_idx, render_dict
                )
        next_idx = num_end + 1
    elif token in AB_Tail_Tokens:
        if l_split[idx + 1] == "{":
            num_start = idx + 1
            num_end = find_matching_brace(l_split, num_start)
            l_split_copy = (
                l_split[:idx]
                + [r"\mathcolor{black}{"]
                + l_split[idx : idx + 2]
                + [r"\mathcolor{gray}{"]
                + l_split[num_start + 1 : num_end]
                + ["}}"]
                + l_split[num_end:]
            )
            l_new = " ".join(l_split_copy)
            l_new = r"\mathcolor{gray}{ " + l_new + " }"
            render_dict[str(idx)] = l_new, token
            sub_idx = num_start + 1
            while sub_idx < num_end:
                l_split, sub_idx, render_dict = token_add_color(
                    l_split, sub_idx, render_dict
                )
            next_idx = num_end + 1
        elif l_split[idx + 1] == "[":
            num_start = idx + 1
            num_end = find_matching_brace(l_split, num_start, brace=["[", "]"])
            den_start = num_end + 1
            den_end = find_matching_brace(l_split, den_start)
            l_split_copy = (
                l_split[:idx]
                + [r"{\mathcolor{black}{"]
                + l_split[idx : idx + 2]
                + [r"\mathcolor{gray}{"]
                + l_split[idx + 2 : num_end]
                + ["}"]
                + l_split[num_end : den_start + 1]
                + [r"\mathcolor{gray}{"]
                + l_split[den_start + 1 : den_end]
                + ["}"]
                + l_split[den_end : den_end + 1]
                + ["}}"]
                + l_split[den_end + 1 :]
            )
            l_new = " ".join(l_split_copy)
            l_new = r"\mathcolor{gray}{ " + l_new + " }"
            render_dict[str(idx)] = l_new, token
            sub_idx = num_start + 1
            while sub_idx < num_end:
                l_split, sub_idx, render_dict = token_add_color(
                    l_split, sub_idx, render_dict
                )
            sub_idx = den_start + 1
            while sub_idx < den_end:
                l_split, sub_idx, render_dict = token_add_color(
                    l_split, sub_idx, render_dict
                )
            next_idx = den_end + 1
    elif token in ["\\multicolumn", "\\multirow"]:
        first_start = idx + 1
        first_end = find_matching_brace(l_split, first_start)
        second_start = first_end + 1
        second_end = find_matching_brace(l_split, second_start)
        third_start = second_end + 1
        third_end = find_matching_brace(l_split, third_start)

        sub_idx = third_start + 1
        while sub_idx < third_end:
            l_split, sub_idx, render_dict = token_add_color(
                l_split, sub_idx, render_dict
            )
        next_idx = third_end + 1
    elif token in SKIP_Tokens + TWO_Tail_Invisb_Tokens or any(
        re.match(pattern, token) for pattern in SKIP_PATTERNS
    ):
        if (token == "[" and l_split[idx - 1] != "\\sqrt") or (
            token == "]" and idx >= 3 and l_split[idx - 3] != "\\sqrt"
        ):
            l_split_copy = l_split.copy()
            l_split_copy[idx] = r"\mathcolor{black}{ " + l_split_copy[idx] + " }"
            l_new = " ".join(l_split_copy)
            l_new = r"\mathcolor{gray}{ " + l_new + " }"
            render_dict[str(idx)] = l_new, token
            next_idx = idx + 1
        else:
            next_idx = idx + 1
    else:
        l_split_copy = l_split.copy()
        l_split_copy[idx] = r"\mathcolor{black}{ " + l_split_copy[idx] + " }"

        l_new = " ".join(l_split_copy)
        l_new = r"\mathcolor{gray}{ " + l_new + " }"
        render_dict[str(idx)] = l_new, token
        next_idx = idx + 1

    return l_split, next_idx, render_dict


def token_add_color_RGB(l_split, idx, token_list, brace_color=False):
    """using \mathcolor[RGB]{r,g,b} to render latex."""
    token = l_split[idx]
    if not token:
        next_idx = idx + 1
    elif token in PHANTOM_Tokens:
        if l_split[idx + 1] == "{":
            brace_end = find_matching_brace(l_split, idx + 1)
        else:
            brace_end = idx + 1
        next_idx = brace_end + 1
    elif token in TWO_Tail_Tokens:
        num_start = idx + 1
        num_end = find_matching_brace(l_split, num_start)
        den_start = num_end + 1
        den_end = find_matching_brace(l_split, den_start)
        color_token = "\\mathcolor[RGB]{<color_<idx>>}{".replace(
            "<idx>", str(len(token_list))
        )
        l_split = (
            l_split[:idx]
            + [color_token + token]
            + l_split[idx + 1 : den_end + 1]
            + ["}"]
            + l_split[den_end + 1 :]
        )
        token_list.append(token)
        next_idx = idx + 1
    elif token in ONE_Tail_Tokens:
        num_start = idx + 1
        num_end = find_matching_brace(l_split, num_start)
        color_token = "\\mathcolor[RGB]{<color_<idx>>}{".replace(
            "<idx>", str(len(token_list))
        )
        if (
            token != "\\underbrace"
            and num_end + 1 < len(l_split)
            and l_split[num_end + 1] == "_"
        ):
            l_split = (
                l_split[:idx]
                + ["{" + color_token + token]
                + l_split[idx + 1 : num_end + 1]
                + ["}}"]
                + l_split[num_end + 1 :]
            )
        else:
            l_split = (
                l_split[:idx]
                + [color_token + token]
                + l_split[idx + 1 : num_end + 1]
                + ["}"]
                + l_split[num_end + 1 :]
            )
        token_list.append(token)
        next_idx = idx + 1
    elif token in ONE_Tail_Invisb_Tokens:
        num_start = idx + 1
        num_end = find_matching_brace(l_split, num_start)
        sub_idx = num_start + 1
        if num_end - num_start == 2:
            color_token = "\\mathcolor[RGB]{<color_<idx>>}{".replace(
                "<idx>", str(len(token_list))
            )
            token_list.append(l_split[num_start + 1])
            l_split = (
                l_split[: num_start + 1]
                + [color_token + l_split[num_start + 1] + "}"]
                + l_split[num_end:]
            )
        else:
            while sub_idx < num_end:
                l_split, sub_idx, token_list = token_add_color_RGB(
                    l_split, sub_idx, token_list
                )
        next_idx = num_end + 1
    elif token in AB_Tail_Tokens:
        if l_split[idx + 1] == "{":
            num_start = idx + 1
            num_end = find_matching_brace(l_split, num_start)
            color_token = "\\mathcolor[RGB]{<color_<idx>>}{".replace(
                "<idx>", str(len(token_list))
            )
            l_split = (
                l_split[:idx]
                + [color_token + token]
                + l_split[idx + 1 : num_end + 1]
                + ["}"]
                + l_split[num_end + 1 :]
            )
            token_list.append(token)
            sub_idx = num_start + 1
            while sub_idx < num_end:
                l_split, sub_idx, token_list = token_add_color_RGB(
                    l_split, sub_idx, token_list
                )
            next_idx = num_end + 1
        elif l_split[idx + 1] == "[":
            num_start = idx + 1
            num_end = find_matching_brace(l_split, num_start, brace=["[", "]"])
            den_start = num_end + 1
            den_end = find_matching_brace(l_split, den_start)
            color_token = "\\mathcolor[RGB]{<color_<idx>>}{".replace(
                "<idx>", str(len(token_list))
            )
            l_split = (
                l_split[:idx]
                + [color_token + token]
                + l_split[idx + 1 : den_end + 1]
                + ["}"]
                + l_split[den_end + 1 :]
            )
            token_list.append(token)
            sub_idx = num_start + 1
            while sub_idx < num_end:
                l_split, sub_idx, token_list = token_add_color_RGB(
                    l_split, sub_idx, token_list, brace_color=True
                )
            sub_idx = den_start + 1
            while sub_idx < den_end:
                l_split, sub_idx, token_list = token_add_color_RGB(
                    l_split, sub_idx, token_list
                )
            next_idx = den_end + 1
    elif token in ["\\multicolumn", "\\multirow"]:
        first_start = idx + 1
        first_end = find_matching_brace(l_split, first_start)
        second_start = first_end + 1
        second_end = find_matching_brace(l_split, second_start)
        third_start = second_end + 1
        third_end = find_matching_brace(l_split, third_start)

        sub_idx = third_start + 1
        while sub_idx < third_end:
            l_split, sub_idx, token_list = token_add_color_RGB(
                l_split, sub_idx, token_list
            )
        next_idx = third_end + 1
    elif token in SKIP_Tokens + TWO_Tail_Invisb_Tokens or any(
        re.match(pattern, token) for pattern in SKIP_PATTERNS
    ):

        if (token == "[" and l_split[idx - 1] != "\\sqrt") or (
            token == "]" and idx >= 3 and l_split[idx - 3] != "\\sqrt"
        ):
            color_token = "\\mathcolor[RGB]{<color_<idx>>}{".replace(
                "<idx>", str(len(token_list))
            )
            l_split = (
                l_split[:idx] + [color_token + l_split[idx] + "}"] + l_split[idx + 1 :]
            )
            token_list.append(token)
            next_idx = idx + 1
        else:
            next_idx = idx + 1
    else:

        if brace_color or (idx > 1 and l_split[idx - 1] == "_"):
            color_token = "\\mathcolor[RGB]{<color_<idx>>}{".replace(
                "<idx>", str(len(token_list))
            )
            l_split = (
                l_split[:idx]
                + ["{" + color_token + l_split[idx] + "}}"]
                + l_split[idx + 1 :]
            )
            token_list.append(token)
            next_idx = idx + 1
        else:
            color_token = "\\mathcolor[RGB]{<color_<idx>>}{".replace(
                "<idx>", str(len(token_list))
            )
            l_split = (
                l_split[:idx] + [color_token + l_split[idx] + "}"] + l_split[idx + 1 :]
            )
            token_list.append(token)
            next_idx = idx + 1
    return l_split, next_idx, token_list



import re

# 移除 pred 字段首尾的 \[ 和 \]
# 该正则用于去掉字符串开头被颜色宏包裹的 \[，确保公式起始干净
PATTERN_STRIP_START_BRACKET = re.compile(
    r"^\s*\\color\{#[\da-fA-F]{6}\}\{\s*\\\[\s*\}\s*"
)
# 该正则用于去掉字符串末尾被颜色宏包裹的 \]，确保公式结尾干净
PATTERN_STRIP_END_BRACKET = re.compile(
    r"\s*\\color\{#[\da-fA-F]{6}\}\{\s*\\\]\s*\}\s*$"
)


def full_to_half_width(s: str) -> str:
    """
    将字符串中的全角字符转换为半角字符。
    """
    res = ""
    for char in s:
        inside_code = ord(char)
        if inside_code == 12288:  # 全角空格
            inside_code = 32
        elif 65281 <= inside_code <= 65374:  # 其他全角字符
            inside_code -= 65248
        res += chr(inside_code)
    return res


def clean_latex_delimiters(latex_string: str) -> str:
    """
    清理 LaTeX 字符串中的美元符号和双美元符号。
    """
    s = latex_string.strip()
    if s.startswith("$$") and s.endswith("$$"):
        return s[2:-2].strip()
    if s.startswith("$") and s.endswith("$"):
        return s[1:-1].strip()
    return s


def clean(s: str) -> str:
    """
    对字符串进行预处理，包括全角转半角、去空格等。
    """
    processed_text = full_to_half_width(s)
    processed_text = processed_text.replace("{/[}", r" \[").replace("{/]}", r" \]")
    cleaned_text = clean_latex_delimiters(processed_text)
    return cleaned_text



from fastcdm.render.render_worker import RenderWorker
from fastcdm.matcher import update_inliers, HungarianMatcher, SimpleAffineTransform
from fastcdm.clean import (
    clean,
    PATTERN_STRIP_START_BRACKET,
    PATTERN_STRIP_END_BRACKET,
)
from fastcdm.tokenize import tokenize
from fastcdm.colorize import process_for_katex, generate_high_contrast_colors
from fastcdm.box import get_bboxes_from_array

import cv2
import numpy as np
from typing import List, Tuple
from pathlib import Path
from skimage.measure import ransac
import traceback

root_dir = Path(__file__).parent
TEMPLATE_FILE = root_dir / "render" / "templates" / "formula.html"


def preprocess(s: str):
    # --- Step 1: Clean & Tokenization ---
    clean_s = clean(s)
    success_tokenization, tokenized_s = tokenize(clean_s)

    if not success_tokenization:
        print("Tokenization failed")
        return 0.0

    # --- Step 2: Prepare Colorized Latex for KaTeX ---
    katex_template, token_list = process_for_katex(tokenized_s)

    # Generate colors
    num_colors = len(token_list) + 10
    colors_rgb = generate_high_contrast_colors(num_colors)

    final_latex = katex_template
    color_map = []  # List of (token, rgb_color)
    for c_idx, token in enumerate(token_list):
        r, g, b = colors_rgb[c_idx % len(colors_rgb)]
        final_latex = final_latex.replace(
            f"__COLOR__{c_idx}__", f"#{r:02x}{g:02x}{b:02x}"
        )
        color_map.append((token, (r, g, b)))

    # 移除首尾括号
    final_latex = PATTERN_STRIP_START_BRACKET.sub("", final_latex)
    final_latex = PATTERN_STRIP_END_BRACKET.sub("", final_latex)

    return final_latex, color_map


def calculate_metrics(gt_len, pred_len, match_num):
    """计算F1-score, Recall, Precision。"""
    recall = match_num / gt_len if gt_len > 0 else 0
    precision = match_num / pred_len if pred_len > 0 else 0
    f1_score = (
        2 * (precision * recall) / (precision + recall) if recall + precision > 0 else 0
    )
    return f1_score, recall, precision


def postprocess(
    img_gt: np.ndarray,
    img_pred: np.ndarray,
    gt_color_map: List[Tuple[str, Tuple[int, int, int]]],
    pred_color_map: List[Tuple[str, Tuple[int, int, int]]],
    visualize: bool,
):
    # Normalize Image Sizes (Max Height/Width)
    h_gt, w_gt = img_gt.shape[:2]
    h_pred, w_pred = img_pred.shape[:2]
    max_h = max(h_gt, h_pred)
    max_w = max(w_gt, w_pred)

    # Create canvas
    final_gt_img = np.full((max_h, max_w, 3), 255, dtype=np.uint8)
    final_gt_img[0:h_gt, 0:w_gt] = img_gt

    final_pred_img = np.full((max_h, max_w, 3), 255, dtype=np.uint8)
    final_pred_img[0:h_pred, 0:w_pred] = img_pred

    vis_img = None

    gt_colors = [c[1] for c in gt_color_map]
    pred_colors = [c[1] for c in pred_color_map]

    # bboxes format: [xmin, ymin, xmax, ymax]
    gt_bboxes_list = get_bboxes_from_array(final_gt_img, gt_colors)
    pred_bboxes_list = get_bboxes_from_array(final_pred_img, pred_colors)

    # --- Step 3: Match Tokens ---
    # Convert to list of dicts as expected by HungarianMatcher
    # item = {'bbox': [xmin, ymin, xmax, ymax], 'token': token_str}

    gt_data = []
    for i, bbox in enumerate(gt_bboxes_list):
        if bbox:
            gt_data.append({"bbox": bbox, "token": gt_color_map[i][0]})

    pred_data = []
    for i, bbox in enumerate(pred_bboxes_list):
        if bbox:
            pred_data.append({"bbox": bbox, "token": pred_color_map[i][0]})

    matcher = HungarianMatcher()
    size_tuple = (max_w, max_h)

    matched_idxes = matcher(gt_data, pred_data, size_tuple, size_tuple)

    # RANSAC Verification
    src, dst = [], []
    for idx1, idx2 in matched_idxes:
        # Center points
        x1_c = (gt_data[idx1]["bbox"][0] + gt_data[idx1]["bbox"][2]) / 2
        y1_c = (gt_data[idx1]["bbox"][1] + gt_data[idx1]["bbox"][3]) / 2
        x2_c = (pred_data[idx2]["bbox"][0] + pred_data[idx2]["bbox"][2]) / 2
        y2_c = (pred_data[idx2]["bbox"][1] + pred_data[idx2]["bbox"][3]) / 2
        src.append([y1_c, x1_c])
        dst.append([y2_c, x2_c])

    src, dst = np.array(src), np.array(dst)
    min_samples = 3

    if src.shape[0] <= min_samples:
        inliers = np.ones(len(matched_idxes), dtype=bool)
    else:
        inliers = np.zeros(len(matched_idxes), dtype=bool)
        for _ in range(5):
            if np.sum(~inliers) <= min_samples:
                break
            # SimpleAffineTransform expects (N, 2)
            # RANSAC fits model to data
            try:
                model, inliers_1 = ransac(
                    (src[~inliers], dst[~inliers]),
                    SimpleAffineTransform,
                    min_samples=min_samples,
                    residual_threshold=20,
                    max_trials=50,
                )
                if inliers_1 is not None and inliers_1.any():
                    inliers = update_inliers(inliers, inliers_1)
                else:
                    break
            except Exception:
                # Ransac might fail if data is degenerate
                break

    # Double check token cost for inliers
    for idx, (a, b) in enumerate(matched_idxes):
        # matcher.cost['token'] is (gt_len, pred_len)
        # If token cost is 1 (completely different), reject even if spatially aligned?
        # visual_matcher.py logic: if inliers[idx] and matcher.cost['token'][a, b] == 1: inliers[idx] = False
        if inliers[idx] and matcher.cost["token"][a, b] == 1:
            inliers[idx] = False

    match_num = np.sum(inliers)

    num_gt = len(gt_bboxes_list)
    num_pred = len(pred_bboxes_list)

    f1, recall, precision = calculate_metrics(num_gt, num_pred, match_num)

    if visualize:
        vis_img = np.full((max_h * 2 + 10, max_w, 3), 255, dtype=np.uint8)
        vis_img[0:max_h, 0:max_w] = final_gt_img
        vis_img[max_h + 10 : max_h + 10 + max_h, 0:max_w] = final_pred_img

        # Draw matches that are inliers
        for idx, (gt_idx, pred_idx) in enumerate(matched_idxes):
            if inliers[idx]:
                gt_box = gt_data[gt_idx]["bbox"]
                pred_box = pred_data[pred_idx]["bbox"]

                # Draw boxes
                cv2.rectangle(
                    vis_img,
                    (gt_box[0], gt_box[1]),
                    (gt_box[2], gt_box[3]),
                    (0, 255, 0),
                    1,
                )
                y_offset = max_h + 10
                cv2.rectangle(
                    vis_img,
                    (pred_box[0], pred_box[1] + y_offset),
                    (pred_box[2], pred_box[3] + y_offset),
                    (0, 0, 255),
                    1,
                )

                # Draw line
                pt1 = (
                    int((gt_box[0] + gt_box[2]) / 2),
                    int((gt_box[1] + gt_box[3]) / 2),
                )
                pt2 = (
                    int((pred_box[0] + pred_box[2]) / 2),
                    int((pred_box[1] + pred_box[3]) / 2) + y_offset,
                )
                cv2.line(vis_img, pt1, pt2, (255, 0, 0), 1)

    return (f1, recall, precision, vis_img) if visualize else (f1, recall, precision)


class FastCDM:
    def __init__(self, chromedriver: str = None) -> None:
        self.chromedriver = chromedriver
        self.render_worker = None
        self.init_render_worker()

    def init_render_worker(self) -> None:
        try:
            self.render_worker = RenderWorker(
                template_file="file://" + str(TEMPLATE_FILE.resolve()),
                timeout=30,
                driver_path=self.chromedriver,
            )
        except Exception as e:
            print("Failed to init RenderWorker:")
            print("=" * 30)
            print(traceback.format_exc(e))
            return None

    def render(self, latex_list: list) -> list:
        """
        渲染 LaTeX 表达式列表。

        参数:
            latex_list (list): LaTeX 表达式列表。

        返回:
            list: 渲染后的图像列表。
        """
        try:
            latex_strings = [
                f"$${s}$$" if not s.startswith("$$") else s for s in latex_list
            ]
            imgs = self.render_worker.render(latex_strings)
        except Exception as e:
            print("Rendering failed:")
            print("=" * 30)
            print(traceback.format_exc(e))
            return []

        assert len(imgs) == len(
            latex_strings
        ), "Number of rendered images must match number of input strings"
        return imgs

    def compute(self, gt: str, pred: str, visualize: bool = False) -> tuple:
        """
        计算给定的 GT 和预测 LaTeX 表达式的 CDM 分数。

        参数:
            gt (str):  ground truth LaTeX 表达式。
            pred (str): 预测 LaTeX 表达式。

        返回:
            tuple: 包含 F1 分数、召回率和准确率的元组。
        """
        gt_latex, gt_color_map = preprocess(gt)
        pred_latex, pred_color_map = preprocess(pred)

        imgs = self.render([gt_latex, pred_latex])
        gt_img, pred_img = imgs[0], imgs[1]

        result = postprocess(gt_img, pred_img, gt_color_map, pred_color_map, visualize)
        return result

    def batch_compute(self, gt_list: list, pred_list: list) -> list:
        """
        TODO
        批量计算给定的 GT 和预测 LaTeX 表达式的 CDM 分数。

        参数:
            gt_list (list):  ground truth LaTeX 表达式列表。
            pred_list (list): 预测 LaTeX 表达式列表。

        返回:
            list: 包含每个表达式的 F1 分数、召回率和准确率的元组列表。
        """
        raise NotImplementedError("batch_compute is not implemented yet.")


if __name__ == "__main__":
    # gt = r"A_{M123}=u\,A^{M}"
    # pred = r"A_M123 = \hat{u} A^M"

    # gt = r"r = \frac { \alpha } { \beta } \vert \sin \beta \left( \sigma _ { 1 } \pm \sigma _ { 2 } \right) \vert"
    # pred = r"r={\frac{\alpha}{\beta}}|\sin\beta\left(\sigma_{2}+\sigma_{1}\right)|"

    # gt = r"\frac{1}{2}"
    # pred = r"\frac{1}{2}"

    # gt = r"\tilde{\theta}_k(t)=\frac{\hat{\theta}_k(t+1)-\hat{\theta}_k(t)}{T_s}"
    # pred = r"\tilde{\theta}_k(t)=\frac{\hat{\theta}_k(t+1)-\hat{\theta}_k(t)}{T_s}"

    gt = r"\begin{bmatrix}(\mathbf{I}-\mathbf{A}^{\mathsf{DD}})&-\mathbf{A}^{\mathsf{DP }}&-\mathbf{A}^{\mathsf{DN}}\\ 0&\mathbf{I}&0\\ -\mathbf{A}^{\mathsf{ND}}&-\mathbf{A}^{\mathsf{NP}}&(\mathbf{I}-\mathbf{A}^{ \mathsf{NN}})\end{bmatrix}^{-1}=\begin{bmatrix}\mathbf{B}^{\mathsf{DD}}& \mathbf{B}^{\mathsf{DP}}&\mathbf{B}^{\mathsf{DN}}\\ \mathbf{B}^{\mathsf{PD}}&\mathbf{B}^{\mathsf{PP}}&\mathbf{B}^{\mathsf{PN}}\\ \mathbf{B}^{\mathsf{ND}}&\mathbf{B}^{\mathsf{NP}}&\mathbf{B}^{\mathsf{NN}} \end{bmatrix}"
    pred = r"\left[ \begin{array} { c c c } { ( I - A ^ { \mathrm { D D } } ) } & { - A ^ { \mathrm { D P } } } & { - A ^ { \mathrm { D N } } } \\ { 0 } & { \mathbf { I } } & { 0 } \\ { - A ^ { \mathrm { N D } } } & { - A ^ { \mathrm { N P } } } & { ( I - A ^ { \mathrm { N N } } ) } \end{array} \right] ^ { - 1 } = \left[ \begin{array} { c c c } { \mathbf { B } ^ { \mathrm { D D } } } & { \mathbf { B } ^ { \mathrm { D P } } } & { \mathbf { B } ^ { \mathrm { D N } } } \\ { \mathbf { B } ^ { \mathrm { P D } } } & { \mathbf { B } ^ { \mathrm { P P } } } & { \mathbf { B } ^ { \mathrm { P N } } } \\ { \mathbf { B } ^ { \mathrm { N D } } } & { \mathbf { B } ^ { \mathrm { N P } } } & { \mathbf { B } ^ { \mathrm { N N } } } \end{array} \right]"

    fastcdm = FastCDM(chromedriver="driver/chromedriver")
    res = fastcdm.compute(gt, pred, visualize=True)
    f1, recall, precision, vis_img = res
    print(f"CDM Score (F1): {f1:.4f}")
    print(f"Recall: {recall:.4f}")
    print(f"Precision: {precision:.4f}")
    try:
        out_dir = Path(__file__).parent.parent / "vis"
        out_dir.mkdir(parents=True, exist_ok=True)
        cv2.imwrite(str(out_dir / "match_vis.png"), vis_img)
    except Exception:
        pass



from fastcdm.core import preprocess

def test_preprocess():
    latex_string = r"\begin{bmatrix}(\mathbf{I}-\mathbf{A}^{\mathsf{DD}})&-\mathbf{A}^{\mathsf{DP }}&-\mathbf{A}^{\mathsf{DN}}\\ 0&\mathbf{I}&0\\ -\mathbf{A}^{\mathsf{ND}}&-\mathbf{A}^{\mathsf{NP}}&(\mathbf{I}-\mathbf{A}^{ \mathsf{NN}})\end{bmatrix}^{-1}=\begin{bmatrix}\mathbf{B}^{\mathsf{DD}}& \mathbf{B}^{\mathsf{DP}}&\mathbf{B}^{\mathsf{DN}}\\ \mathbf{B}^{\mathsf{PD}}&\mathbf{B}^{\mathsf{PP}}&\mathbf{B}^{\mathsf{PN}}\\ \mathbf{B}^{\mathsf{ND}}&\mathbf{B}^{\mathsf{NP}}&\mathbf{B}^{\mathsf{NN}} \end{bmatrix}"
    processed, _ = preprocess(latex_string)
    print(processed)

test_preprocess()



import os
import cv2
import random
import numpy as np
from typing import List

from selenium import webdriver
from selenium.webdriver.chrome.service import Service as ChromeService
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from webdriver_manager.chrome import ChromeDriverManager


class RenderWorker:
    """
    一个使用 Selenium Headless Chrome 渲染HTML内容的工具类。
    它可以加载一个HTML模板，通过JavaScript渲染内容（如数学公式），
    并截取渲染后各元素的图像。
    """

    def __init__(self, template_file: str, timeout: int = 15, driver_path: str = None):
        # --- 配置浏览器选项 ---
        opts = Options()
        opts.add_argument("--headless")
        opts.add_argument("--no-sandbox")
        opts.add_argument("--disable-gpu")
        opts.add_argument("--hide-scrollbars")
        opts.add_argument("--disable-dev-shm-usage")
        opts.add_argument("--log-level=3")
        opts.add_experimental_option("excludeSwitches", ["enable-logging"])
        opts.add_argument("--disable-font-antialiasing")
        opts.add_argument("--allow-file-access-from-files")

        # --- 使用 webdriver-manager 自动管理 ChromeDriver（需要代理下载）---
        if driver_path is None:
            driver_path = ChromeDriverManager().install()
            print(f"Installed driver_path: {driver_path}")
        elif not os.path.exists(driver_path):
            # TODO：如果指定路径不存在，则在该路径安装 ChromeDriver
            raise FileNotFoundError(f"ChromeDriver 未找到：{driver_path}")

        # 临时清除代理环境变量，避免 Selenium 连接到 ChromeDriver 时出现问题
        saved_http_proxy = os.environ.pop("http_proxy", None)
        saved_https_proxy = os.environ.pop("https_proxy", None)

        try:
            service = ChromeService(driver_path)
            # --- 初始化 WebDriver ---
            self.driver = webdriver.Chrome(service=service, options=opts)
        finally:
            # 恢复代理设置
            if saved_http_proxy:
                os.environ["http_proxy"] = saved_http_proxy
            if saved_https_proxy:
                os.environ["https_proxy"] = saved_https_proxy

        self.driver.get(template_file)

        self.timeout = timeout

        # 定义一个固定的窗口宽度
        self.window_fix_width = 2000
        self.window_init_height = 300

        self.driver.set_window_size(self.window_fix_width, self.window_init_height)

        # 动态计算浏览器边框和工具栏的高度
        self.outer_height = self.window_init_height - self.driver.execute_script(
            "return window.innerHeight"
        )

        # 等待页面容器加载完成
        WebDriverWait(self.driver, self.timeout).until(
            EC.presence_of_all_elements_located((By.ID, "container"))
        )

    def render(self, contents: List[str]) -> List[np.ndarray]:
        """
        渲染一组内容并返回每个元素的截图。
        """
        # 通过JS调用页面内的render函数
        self.driver.execute_script(
            "document.body.classList.remove('rendering-complete');"
        )
        self.driver.execute_script(f"render({contents}, false)")

        # 等待JS渲染完成的信号
        WebDriverWait(self.driver, self.timeout).until(
            EC.presence_of_element_located((By.CLASS_NAME, "rendering-complete"))
        )

        # 根据内容的总高度调整窗口大小，以确保能截取完整图像
        scroll_height = self.driver.execute_script(
            "return document.getElementById('container').scrollHeight"
        )
        # Chrome 窗口高度限制（通常最大约为 2^31 像素，但实际会更小）
        # 设置一个安全的最大高度值
        MAX_WINDOW_HEIGHT = 10000
        # 确保 outer_height 不会导致负值或过小值
        # 使用绝对值并添加额外的边距以确保内容完全可见
        safe_outer_height = max(abs(self.outer_height), 100)
        target_height = min(
            max(scroll_height + safe_outer_height, 100), MAX_WINDOW_HEIGHT
        )
        self.driver.set_window_size(self.window_fix_width, target_height)

        # 获取整个页面的截图
        png = self.driver.get_screenshot_as_png()
        nparr = np.frombuffer(png, np.uint8)
        fullpage_img = cv2.imdecode(nparr, cv2.IMREAD_COLOR)

        # 获取每个渲染元素的边界框
        rects = self.get_rects()
        cropped_imgs = []
        img_h, img_w = fullpage_img.shape[:2]

        # 根据边界框裁剪出每个元素的图像
        for rect in rects:
            if rect is None:
                cropped_imgs.append(None)
            else:
                x, y, w, h = rect
                # 计算一个小的随机边距，让截图更自然
                max_side = max(w, h)
                base_border = int(max_side * 0.03)
                border_size = int(base_border * random.uniform(0.8, 1.2))
                x1 = max(0, x - border_size)
                y1 = max(0, y - border_size)
                x2 = min(img_w, x + w + border_size)
                y2 = min(img_h, y + h + border_size)

                cropped = fullpage_img[y1:y2, x1:x2]
                cropped_imgs.append(cropped)

        return cropped_imgs

    def get_rects(self) -> list:
        """
        获取页面上所有渲染元素的位置和大小信息。
        """
        elements = WebDriverWait(self.driver, self.timeout).until(
            EC.presence_of_all_elements_located((By.CLASS_NAME, "screenshot"))
        )

        rects = []
        for element in elements:
            location = element.location
            size = element.size
            x = int(location["x"])
            y = int(location["y"])
            w = int(size["width"])
            h = int(size["height"])

            # 如果元素宽度超过窗口，可能是一个渲染错误，标记为None
            if w > self.window_fix_width:
                rects.append(None)
            else:
                rects.append((x, y, w, h))

        return rects

    def close(self):
        """
        关闭浏览器驱动并释放资源。
        """
        if self.driver:
            self.driver.quit()
            self.driver = None

    def __del__(self):
        """
        对象销毁时确保浏览器被关闭。
        """
        try:
            self.close()
        except:
            pass



import numpy as np
import cv2


def get_bboxes_from_array(img_array, colors):
    bboxes = []
    for color in colors:  # color is RGB
        # img_array is BGR (cv2 default)
        r, g, b = color
        target_bgr = np.array([b, g, r], dtype=np.uint8)

        # Create mask
        mask = cv2.inRange(img_array, target_bgr, target_bgr)
        coords = np.argwhere(mask)

        if coords.size > 0:
            y_min, x_min = coords.min(axis=0)
            y_max, x_max = coords.max(axis=0)
            # [xmin, ymin, xmax, ymax]
            bboxes.append([int(x_min), int(y_min), int(x_max), int(y_max)])
        else:
            bboxes.append([])  # Empty list for missing token
    return bboxes
